from collections import namedtuple

import numpy as np


class FitData:
    _Fits = namedtuple("Fits", ("horizontal", "vertical"))
    _Data = namedtuple("Data", ("x", "y"))
    _Parameters = namedtuple("Parameters", ["offset", "slope", "amplitude", "mean", "sigma", "center", "rms", "peak"])

    def __init__(self):
        self._fits = None
        self._parameters = None

    @property
    def fits(self):
        return self._fits

    @property
    def parameters(self):
        return self._parameters

    def update(self, get_gaussian_fit, projections_values_horizontal, projections_values_vertical):
        horizontal_fit, horizontal_fit_parameters = get_gaussian_fit(projections_values_horizontal)
        vertical_fit, vertical_fit_parameters = get_gaussian_fit(projections_values_vertical)
        self._fits = self._Fits(self._unpack_fit(horizontal_fit), self._unpack_fit(vertical_fit))
        self._parameters = self._Fits(
            self._unpack_fit_parameters(horizontal_fit_parameters, projections_values_horizontal),
            self._unpack_fit_parameters(vertical_fit_parameters, projections_values_vertical))

    def _unpack_fit(self, fit):
        if fit is None:
            return None
        return self._Data(*fit)

    def _unpack_fit_parameters(self, fit_parameters, projections_values):
        if fit_parameters is None:
            return self._Parameters(np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan)
        return self._Parameters(*self._get_parameters(fit_parameters, projections_values))

    @staticmethod
    def _get_parameters(fit_parameters, projections_values):
        y_corrected = abs(projections_values.y - fit_parameters[0] - fit_parameters[1] *
                          (projections_values.x - projections_values.x[0]))

        mean = fit_parameters[3]
        sigma = fit_parameters[4]
        mean_area = abs(projections_values.x - mean) <= 5 * sigma

        center = np.nan
        rms = np.nan
        peak = np.nan
        if any(mean_area):
            center = (np.nansum(projections_values.x[mean_area] * y_corrected[mean_area]) /
                      np.nansum(y_corrected[mean_area]))

            rms = np.sqrt(np.nansum((projections_values.x[mean_area] - center) ** 2 * y_corrected[mean_area]) /
                          np.nansum(y_corrected[mean_area]))

            peak = projections_values.x[mean_area][np.nanargmax(y_corrected[mean_area])]
        return *fit_parameters, center, rms, peak
