from collections import namedtuple

import numpy as np
from scipy import stats
from scipy.signal import find_peaks


class ReflectivityData:
    _Signal = namedtuple("Signal", ["bins", "values"])
    _Reflectivity = namedtuple("Reflectivity", ("values", "errors"))

    class UtilizedNumberOfPulsesBelowMinimumAcceptable(Exception):
        def __init__(self):
            super().__init__("Utilized number of pulses below minimum acceptable (N = 2)")

    def __init__(self, adc_address, bunch_number_address):
        self._adc_address = adc_address
        self._bunch_number_address = bunch_number_address
        self._use_fast_shutter = None
        self._adc = None
        self._threshold = None
        self._positions = None
        self._laser_pulses = None
        self._fel_pulses = None
        self._utilized_pulses = None
        self._minimum_threshold_factor = None
        self._maximum_threshold_factor = None
        self._signal_range = None
        self._background_range = None
        self._reflectivities = []
        self._reflectivity = None
        self._fast_shutter_signals = []

    @property
    def adc_address(self):
        return self._adc_address

    @property
    def bunch_number_address(self):
        return self._bunch_number_address

    @property
    def adc(self):
        return self._adc

    @property
    def threshold(self):
        return self._threshold

    @property
    def positions(self):
        return self._positions

    @property
    def laser_pulses(self):
        return self._laser_pulses

    @property
    def fel_pulses(self):
        return self._fel_pulses

    @property
    def utilized_pulses(self):
        return self._utilized_pulses

    @property
    def minimum_threshold_factor(self):
        return self._minimum_threshold_factor

    @property
    def maximum_threshold_factor(self):
        return self._maximum_threshold_factor

    @property
    def reflectivity(self):
        return self._reflectivity

    def use_fast_shutter(self, state):
        self._use_fast_shutter = state

    def is_fast_shutter_used(self):
        return self._use_fast_shutter

    def update_pulses(self, adc_trace_of_laser_pulses, number_of_fel_pulses):
        self._adc = self._Signal(np.arange(len(adc_trace_of_laser_pulses)), adc_trace_of_laser_pulses)
        self._fel_pulses = number_of_fel_pulses

    def initialize(self):
        self._reflectivities.clear()
        self._reflectivity = None
        self._fast_shutter_signals.clear()

    def find_threshold(self, factor):
        mode = stats.mode(self._adc.values, keepdims=False)[0]
        iqr = np.diff(np.quantile(self._adc.values, q=[.25, .75]))[0]
        threshold = float(factor) * iqr
        self._minimum_threshold_factor = 2
        try:
            self._maximum_threshold_factor = int((np.nanmax(self._adc.values) - mode) / iqr)
        except ValueError:
            self._maximum_threshold_factor = self._minimum_threshold_factor
        self._threshold = self._Signal(self._adc.bins, np.ones(len(self._adc.bins)) * (mode + threshold))

    def find_pulses(self):
        peaks, _ = find_peaks(self._adc.values, height=self._threshold.values[0])
        self._positions = self._Signal(peaks, self._adc.values[peaks])
        self._laser_pulses = len(peaks)
        if self._use_fast_shutter:
            self._utilized_pulses = self._fel_pulses
            if self._utilized_pulses > self._laser_pulses:
                self._utilized_pulses = self._laser_pulses
        else:
            self._utilized_pulses = 2 * self._fel_pulses
            if self._utilized_pulses > self._laser_pulses:
                self._utilized_pulses = int(self._laser_pulses / 2) * 2

    def set_parameters(self, signal_range, background_range):
        self._signal_range = signal_range
        self._background_range = background_range

    def is_not_empty(self):
        return self._reflectivity is not None

    def update(self, reflectivity_data):
        corrected_signal = np.zeros(self._utilized_pulses)
        for index, peak in enumerate(self._positions.bins[:self._utilized_pulses]):
            background = np.nanmean(reflectivity_data[peak + self._background_range[0]:
                                                      peak + self._background_range[1] + 1])
            corrected_signal[index] = np.nansum((reflectivity_data - background)[peak + self._signal_range[0]:
                                                                                 peak + self._signal_range[1] + 1])
        if self._use_fast_shutter:
            self._fast_shutter_signals.append(np.nansum(corrected_signal))
            if len(self._fast_shutter_signals) == 2:
                reflectivity = self._fast_shutter_signals[0] / self._fast_shutter_signals[1] * 100
                self._reflectivities.extend((reflectivity, reflectivity))
                self._fast_shutter_signals.clear()
        else:
            self._reflectivities.append(np.nansum(corrected_signal[::2]) / np.nansum(corrected_signal[1::2]) * 100)

    def process_with(self, process_function):
        reflectivity = process_function(self._reflectivities)
        self._reflectivity = self._Reflectivity(reflectivity[:, 0], reflectivity[:, 1])
