from collections import namedtuple

import numpy as np
from scipy import stats
from scipy.signal import find_peaks


class TransmissionData:
    _Signal = namedtuple("Signal", ["bins", "values"])
    _Transmission = namedtuple("Transmission", ("values", "errors"))

    def __init__(self, adc_address):
        self._adc_address = adc_address
        self._adc = None
        self._threshold = None
        self._positions = None
        self._pulses = None
        self._minimum_threshold_factor = None
        self._maximum_threshold_factor = None
        self._actual_pulse = None
        self._signal_range = None
        self._background_range = None
        self._transmissions = []
        self._transmission_values = np.array([])
        self._transmission_errors = np.array([])
        self._transmission = None

    @property
    def adc_address(self):
        return self._adc_address

    @property
    def adc(self):
        return self._adc

    @property
    def threshold(self):
        return self._threshold

    @property
    def positions(self):
        return self._positions

    @property
    def pulses(self):
        return self._pulses

    @property
    def minimum_threshold_factor(self):
        return self._minimum_threshold_factor

    @property
    def maximum_threshold_factor(self):
        return self._maximum_threshold_factor

    @property
    def transmission(self):
        return self._transmission

    @property
    def actual_pulse(self):
        return self._actual_pulse

    def initialize(self):
        self._transmissions.clear()
        self._transmission_values = np.array([])
        self._transmission_errors = np.array([])
        self._transmission = None

    def update_pulses(self, adc_trace_of_pulses):
        self._adc = self._Signal(np.arange(len(adc_trace_of_pulses)), adc_trace_of_pulses)

    def find_threshold(self, factor):
        mode = stats.mode(self._adc.values, keepdims=False)[0]
        iqr = np.diff(np.quantile(self._adc.values, q=[.25, .75]))[0]
        threshold = float(factor) * iqr
        self._minimum_threshold_factor = 2
        try:
            self._maximum_threshold_factor = int((np.nanmax(self._adc.values) - mode) / iqr)
        except ValueError:
            self._maximum_threshold_factor = self._minimum_threshold_factor
        self._threshold = self._Signal(self._adc.bins, np.ones(len(self._adc.bins)) * (mode + threshold))

    def find_pulses(self):
        peaks, _ = find_peaks(self._adc.values, height=self._threshold.values[0])
        self._positions = self._Signal(peaks, self._adc.values[peaks])
        self._pulses = len(peaks)

    def set_parameters(self, actual_pulse, signal_range, background_range):
        self._actual_pulse = actual_pulse
        self._signal_range = signal_range
        self._background_range = background_range

    def update(self, transmission_readout_data, profiler_position_data):
        if profiler_position_data.is_ready_for_new_data:
            transmission_data = self._get_transmission_data(transmission_readout_data)
            self._transmissions.append(transmission_data)
            if profiler_position_data.is_ready_for_new_scan_point:
                value, error = self._get_transmissions_value_and_error(self._transmissions,
                                                                       profiler_position_data.scan_step)
                self._transmission_values = np.append(self._transmission_values, value)
                self._transmission_errors = np.append(self._transmission_errors, error)

                self._transmission = self._Transmission(
                    100 * self._transmission_values / self._transmission_values.max(),
                    100 * self._transmission_errors / self._transmission_values.max())

    def _get_transmission_data(self, readout_data):
        peak = self._positions.bins[self._actual_pulse - 1]
        background = np.nanmean(readout_data[peak + self._background_range[0]: peak + self._background_range[1] + 1])
        corrected_signal = np.nansum((readout_data - background)[peak + self._signal_range[0]:
                                                                 peak + self._signal_range[1] + 1])
        return corrected_signal

    @staticmethod
    def _get_transmissions_value_and_error(transmissions, steps):
        last_transmissions = np.array(np.split(np.array(transmissions), steps))[-1]
        return np.nanmean(last_transmissions), np.nanstd(last_transmissions)
