import numpy as np
from doocspie.util import Colors
from pyqtgraph import GraphicsLayoutWidget, ErrorBarItem, PlotDataItem, LegendItem, QtCore, mkPen


class AlignmentPlotter:
    _DESY_ORANGE = np.array(Colors.DESY.ORANGE) * 255
    _DESY_CYAN = np.array(Colors.DESY.CYAN) * 255
    _DESY_GREY = np.array(Colors.DESY.GRAY_1) * 255

    def __init__(self):
        self._widget = GraphicsLayoutWidget()

        self._alignment_plot = self._widget.addPlot()
        self._alignment_error = ErrorBarItem()
        self._alignment_line = PlotDataItem()
        self._alignment_fit = PlotDataItem()

        self._alignment_plot.addItem(self._alignment_error)
        self._alignment_plot.addItem(self._alignment_line)
        self._alignment_plot.addItem(self._alignment_fit)

        self._legend = LegendItem(offset=(70, 365), pen=mkPen(0.5))
        self._current_legend = None

        self._initialize_plots()

    def _initialize_plots(self):
        self._alignment_plot.setMenuEnabled(False)
        self._alignment_plot.showAxis("top", show=True)
        self._alignment_plot.showAxis("right", show=True)
        for pos in "top", "right", "left", "bottom":
            self._alignment_plot.getAxis(pos).setStyle(showValues=False, tickLength=0)
        self._alignment_error.setData(pen={"color": self._DESY_GREY, "width": 2})
        self._alignment_line.setData(symbol="o", pen=None,
                                     symbolBrush=tuple(self._DESY_ORANGE), symbolPen=tuple(self._DESY_ORANGE))
        self._alignment_fit.setData(pen={"color": self._DESY_CYAN, "width": 2})

    @property
    def widget(self):
        return self._widget

    def set_alignment(self, position, metrics_values, metrics_errors):
        self._alignment_error.setData(x=position.values, y=np.array(metrics_values),
                                      width=2 * position.errors, height=2 * np.array(metrics_errors))
        self._alignment_line.setData(x=position.values, y=metrics_values)

        self._alignment_plot.getAxis("left").setStyle(showValues=True)
        self._alignment_plot.setLabel("left", "RMS Width (px)")
        self._alignment_plot.getAxis("bottom").setStyle(showValues=True)
        self._alignment_plot.setLabel("bottom", "Readback Position (mm)")

    def set_fit(self, fit):
        self._alignment_fit.setData(x=fit.position, y=fit.metrics)
        if self._current_legend != fit.legend:
            self._legend.removeItem(self._current_legend)
            self._current_legend = fit.legend
            self._legend.setParentItem(self._alignment_plot)
            self._legend.addItem(self._alignment_fit, fit.legend)
            self._legend.setGeometry(QtCore.QRectF(0, 0, self._legend.width() + 10, self._legend.height()))

    def clear_fit(self):
        self._alignment_fit.clear()
        self._legend.removeItem(self._current_legend)
        self._current_legend = None
        if self._legend.scene():
            self._legend.scene().removeItem(self._legend)

    def clear(self):
        self._alignment_error.setData(x=np.array([]), y=np.array([]), width=np.array([]), height=np.array([]))
        self._alignment_line.clear()
        self._alignment_fit.clear()
        self._alignment_plot.setLabel("left", "")
        self._alignment_plot.getAxis("left").setStyle(showValues=False, tickLength=0)
        self._alignment_plot.setLabel("bottom", "")
        self._alignment_plot.getAxis("bottom").setStyle(showValues=False, tickLength=0)
