import numpy as np
from doocspie.util import Colors
from pyqtgraph import GraphicsLayoutWidget, ErrorBarItem, PlotDataItem, LegendItem, QtCore, mkPen


class ProfilerPlotter:
    _DESY_ORANGE = np.array(Colors.DESY.ORANGE) * 255
    _DESY_CYAN = np.array(Colors.DESY.CYAN) * 255
    _DESY_GREY = np.array(Colors.DESY.GRAY_1) * 255

    def __init__(self):
        self._widget = GraphicsLayoutWidget()

        self._profiler_plot = self._widget.addPlot()
        self._profiler_error = ErrorBarItem()
        self._profiler_line = PlotDataItem()
        self._profiler_fit = PlotDataItem()

        self._profiler_plot.addItem(self._profiler_error)
        self._profiler_plot.addItem(self._profiler_line)
        self._profiler_plot.addItem(self._profiler_fit)

        self._legend = LegendItem(offset=(70, 365), pen=mkPen(0.5))
        self._current_legend = None

        self._initialize_plots()

    def _initialize_plots(self):
        self._profiler_plot.setMenuEnabled(False)
        self._profiler_plot.showAxis("top", show=True)
        self._profiler_plot.showAxis("right", show=True)
        for pos in "top", "right", "left", "bottom":
            self._profiler_plot.getAxis(pos).setStyle(showValues=False, tickLength=0)
        self._profiler_error.setData(pen={"color": self._DESY_GREY, "width": 2})
        self._profiler_line.setData(symbol="o", pen=None,
                                    symbolBrush=tuple(self._DESY_ORANGE), symbolPen=tuple(self._DESY_ORANGE))
        self._profiler_fit.setData(pen={"color": self._DESY_CYAN, "width": 2})

    @property
    def widget(self):
        return self._widget

    def set_profile(self, position, transmission, plot_label):
        self._profiler_error.setData(x=position.values, y=transmission.values,
                                     width=2 * position.errors, height=2 * transmission.errors)
        self._profiler_line.setData(x=position.values, y=transmission.values)

        self._profiler_plot.getAxis("left").setStyle(showValues=True)
        self._profiler_plot.setLabel("left", "Transmission (%)")
        self._profiler_plot.getAxis("bottom").setStyle(showValues=True)
        self._profiler_plot.setLabel("bottom", plot_label)

    def set_fit(self, fit):
        self._profiler_fit.setData(x=fit.position, y=fit.transmission)
        if self._current_legend != fit.legend:
            self._legend.removeItem(self._current_legend)
            self._current_legend = fit.legend
            self._legend.setParentItem(self._profiler_plot)
            self._legend.addItem(self._profiler_fit, fit.legend)
            self._legend.setGeometry(QtCore.QRectF(0, 0, self._legend.width() + 10, self._legend.height()))

    def clear_fit(self):
        self._profiler_fit.clear()
        self._legend.removeItem(self._current_legend)
        self._current_legend = None
        if self._legend.scene():
            self._legend.scene().removeItem(self._legend)

    def clear(self):
        self._profiler_error.setData(x=np.array([]), y=np.array([]), width=np.array([]), height=np.array([]))
        self._profiler_line.clear()
        self._profiler_fit.clear()
        self._profiler_plot.setLabel("left", "")
        self._profiler_plot.getAxis("left").setStyle(showValues=False, tickLength=0)
        self._profiler_plot.setLabel("bottom", "")
        self._profiler_plot.getAxis("bottom").setStyle(showValues=False, tickLength=0)
