import numpy as np
from PyQt5 import QtGui
from doocspie.util import Colors
from pyqtgraph import (setConfigOptions, GraphicsLayoutWidget, PlotDataItem, LabelItem, ImageItem, ROI, QtCore,
                       HistogramLUTItem)


# temporally fix bug in pyqtgraph until it has been fixed in the repo
def stateRect(self, state):
    r = QtCore.QRectF(0, 0, state['size'][0], state['size'][1])
    tr = QtGui.QTransform()
    # tr.rotate(-state['angle'] * 180 / np.pi)
    tr.rotate(state['angle'])
    r = tr.mapRect(r)
    return r.adjusted(state['pos'][0], state['pos'][1], state['pos'][0], state['pos'][1])


ROI.stateRect = stateRect


# -------------------------------------------------------------------


class ViewerPlotter:
    setConfigOptions(imageAxisOrder="row-major")

    _DESY_CYAN = np.array(Colors.DESY.CYAN) * 255

    _WHITESPACE = "&nbsp;"
    _NEWLINE = "<br>"

    def __init__(self):
        self._widget = GraphicsLayoutWidget()

        self._experiment_parameters_label = LabelItem(justify="left")
        self._widget.addItem(self._experiment_parameters_label)

        self._horizontal_projection = PlotDataItem()
        self._horizontal_projection_plot = self._widget.addPlot()
        self._horizontal_projection_plot.addItem(self._horizontal_projection)
        self._horizontal_fit = PlotDataItem()
        self._horizontal_projection_plot.addItem(self._horizontal_fit)

        self._fit_parameters_label = LabelItem(justify="left")
        self._widget.addItem(self._fit_parameters_label)

        self._widget.nextRow()

        self._image = ImageItem()
        self._histogram = HistogramLUTItem()
        self._histogram.setImageItem(self._image)
        self._widget.addItem(self._histogram)

        self._image_plot = self._widget.addPlot()
        self._image_plot.addItem(self._image)

        self._roi = ROI((0, 0), scaleSnap=True, translateSnap=True)
        self._roi_handles = []
        self._image_plot.addItem(self._roi)

        self._vertical_projection = PlotDataItem()
        self._vertical_projection_plot = self._widget.addPlot()
        self._vertical_projection_plot.addItem(self._vertical_projection)
        self._vertical_fit = PlotDataItem()
        self._vertical_projection_plot.addItem(self._vertical_fit)

        self._initialize_plots()
        self._first_readout = True

    def _initialize_plots(self):
        self._initialize_parameters_label()
        self._initialize_horizontal_projection_plot()
        self._initialize_image_plot()
        self._initialize_roi()
        self._initialize_vertical_projection_plot()

    def _initialize_parameters_label(self):
        self._experiment_parameters_label.setMaximumWidth(200)
        self._experiment_parameters_label.setMaximumHeight(250)
        self._fit_parameters_label.setMaximumWidth(200)
        self._fit_parameters_label.setMaximumHeight(250)

    def _initialize_horizontal_projection_plot(self):
        self._initialize_design(self._horizontal_projection_plot)
        self._horizontal_projection_plot.setMouseEnabled(False, False)
        self._horizontal_projection_plot.hideButtons()
        self._horizontal_projection_plot.setMaximumHeight(250)
        self._horizontal_fit.setData(pen={"color": self._DESY_CYAN, "width": 2})

    def _initialize_image_plot(self):
        self._initialize_design(self._image_plot)
        self._image_plot.invertY(True)

    def _initialize_roi(self):
        self._roi.handleSize = 8
        self._add_handles()
        self._roi.setVisible(False)

    def _add_handles(self):
        self._roi_handles.append(self._roi.addScaleHandle((0, 0.5), (0.5, 0.5)))
        self._roi_handles.append(self._roi.addScaleHandle((1, 0.5), (0.5, 0.5)))
        self._roi_handles.append(self._roi.addScaleHandle((0.5, 0), (0.5, 0.5)))
        self._roi_handles.append(self._roi.addScaleHandle((0.5, 1), (0.5, 0.5)))
        self._roi_handles.append(self._roi.addRotateHandle([0, 0], [0.5, 0.5]))
        self._roi_handles.append(self._roi.addRotateHandle([0, 1], [0.5, 0.5]))
        self._roi_handles.append(self._roi.addRotateHandle([1, 0], [0.5, 0.5]))
        self._roi_handles.append(self._roi.addRotateHandle([1, 1], [0.5, 0.5]))

    def _initialize_vertical_projection_plot(self):
        self._initialize_design(self._vertical_projection_plot)
        self._vertical_projection_plot.setMouseEnabled(False, False)
        self._vertical_projection_plot.hideButtons()
        self._vertical_projection_plot.setMaximumWidth(200)
        self._vertical_projection_plot.invertY(True)
        self._vertical_fit.setData(pen={"color": self._DESY_CYAN, "width": 2})

    @staticmethod
    def _initialize_design(plot):
        plot.setMenuEnabled(False)
        plot.showAxis("top", show=True)
        plot.showAxis("right", show=True)
        for position in "top", "right", "left", "bottom":
            plot.getAxis(position).setStyle(showValues=False, tickLength=0)

    @property
    def widget(self):
        return self._widget

    @property
    def roi(self):
        return self._roi

    def add_to_connections_for_changes(self, action):
        self._image_plot.sigRangeChanged.connect(action)
        self._roi.sigRegionChanged.connect(action)

    def get_relevant_image_and_offset(self, image):
        if self._is_image_size_changed(image):
            self._image.setImage(image, autoLevels=self._first_readout)
            self._roi.maxBounds = QtCore.QRectF(0, 0, self._image.width(), self._image.height())
            self.clear_roi()
        return self._roi.getArrayRegion(image, self._image), self._roi.pos()

    def lock_aspect_ratio(self, state):
        self._image_plot.getViewBox().setAspectLocked(state)

    def enable_region_of_interest(self, state):
        self._roi.setVisible(state)

    def _is_image_size_changed(self, image):
        return self._image.height() != image.shape[0] or self._image.width() != image.shape[1]

    def set_image(self, image):
        if self._is_image_size_changed(image):
            self._image.setImage(image, autoLevels=self._first_readout)
            self._roi.maxBounds = QtCore.QRectF(0, 0, self._image.width(), self._image.height())
            self.clear_roi()
        else:
            self._image.setImage(image, autoLevels=self._first_readout)

        if self._first_readout:
            self._histogram.setLevels(image.min(), image.max())
            self._first_readout = False

    def reset(self):
        self._first_readout = True

    def set_roi_position(self, x_position, y_position):
        self._roi.setPos(x_position, y=y_position)

    def set_roi_size(self, width, height):
        self._roi.setSize((width, height))

    def set_roi_rotation(self, angle):
        self._roi.setAngle(angle)

    def clear_roi(self):
        self._roi.setSize([self._image.width(), self._image.height()])
        self._roi.setPos(self._image.pos())
        self._roi.setAngle(0)

    def set_projections(self, values, units):
        self._horizontal_projection.setData(x=values.horizontal.x, y=values.horizontal.y)
        self._horizontal_projection_plot.getAxis("bottom").setStyle(showValues=True)
        self._horizontal_projection_plot.setLabel("bottom", f"Horizontal ({units.horizontal})")

        self._vertical_projection.setData(x=values.vertical.y, y=values.vertical.x)
        self._vertical_projection_plot.getAxis("left").setStyle(showValues=True)
        self._vertical_projection_plot.setLabel("left", f"Vertical ({units.vertical})")

        self.update_axis()

    def set_fits(self, fits):
        if fits.horizontal is None:
            self._horizontal_fit.clear()
        else:
            self._horizontal_fit.setData(x=fits.horizontal.x, y=fits.horizontal.y)

        if fits.vertical is None:
            self._vertical_fit.clear()
        else:
            self._vertical_fit.setData(x=fits.vertical.y, y=fits.vertical.x)

    def set_experiment_parameters(self, average_samples, readback_energy, readback_position):
        self._experiment_parameters_label.setText(self._get_experiment_parameters_label_text(
            average_samples, readback_energy, readback_position))

    def set_fit_parameters(self, fit_parameters):
        self._fit_parameters_label.setText(self._get_fit_parameters_label_text(fit_parameters))

    def _get_experiment_parameters_label_text(self, average_samples, readback_energy, readback_position):
        rows = (f"average samples: {average_samples:4}" + 2 * self._NEWLINE,
                f"readback energy: {readback_energy:4.2f}" + 2 * self._NEWLINE,
                f"readback position: {readback_position:4.2f}" + 2 * self._NEWLINE,
                50 * self._WHITESPACE)  # this realizes padding for the parameters
        return "".join(row for row in rows)

    def _get_fit_parameters_label_text(self, fit_parameters):
        offset = "offset"
        slope = "slope"
        amplitude = "amplitude"
        mean = "mean"
        sigma = "sigma"
        rows = ("horizontal parameters" + self._NEWLINE,
                self._get_parameters_label_row(offset, fit_parameters.horizontal.offset, 1),
                self._get_parameters_label_row(slope, fit_parameters.horizontal.slope, 1),
                self._get_parameters_label_row(amplitude, fit_parameters.horizontal.amplitude, 1),
                self._get_parameters_label_row(mean, fit_parameters.horizontal.mean, 1),
                self._get_parameters_label_row(sigma, fit_parameters.horizontal.sigma, 1),
                self._NEWLINE + 8 * self._WHITESPACE + "vertical parameters" + self._NEWLINE,
                self._get_parameters_label_row(offset, fit_parameters.vertical.offset, 9),
                self._get_parameters_label_row(slope, fit_parameters.vertical.slope, 9),
                self._get_parameters_label_row(amplitude, fit_parameters.vertical.amplitude, 9),
                self._get_parameters_label_row(mean, fit_parameters.vertical.mean, 9),
                self._get_parameters_label_row(sigma, fit_parameters.vertical.sigma, 9))
        return "".join(row for row in rows)

    def _get_parameters_label_row(self, parameter_name, parameter, leading_blanks):
        return leading_blanks * self._WHITESPACE + f"{parameter_name}: {parameter:.2f}" + self._NEWLINE

    def activate_roi(self, state):
        self._roi.translatable = state
        if state:
            if not self._roi_handles:
                self._add_handles()
        else:
            for handle in self._roi_handles:
                self._roi.removeHandle(handle)
            self._roi_handles.clear()

    def update_axis(self):
        if self._roi.isVisible():
            self._horizontal_projection_plot.enableAutoRange()
            self._vertical_projection_plot.enableAutoRange()
        else:
            self._horizontal_projection_plot.setXRange(*self._image_plot.getAxis("bottom").range, padding=0)
            self._vertical_projection_plot.setYRange(*self._image_plot.getAxis("left").range, padding=0)
