import numpy as np


class TimingService:

    def __init__(self, io_service, delay_data, reflectivity_data):
        self._io_service = io_service
        self._delay_data = delay_data
        self._reflectivity_data = reflectivity_data
        self._reflectivity_readout = None
        self._event = None
        self._is_new_data_available = None
        self._last_event_odd = None

    @property
    def delay_data(self):
        return self._delay_data

    @property
    def reflectivity_data(self):
        return self._reflectivity_data

    @property
    def is_new_data_available(self):
        return self._is_new_data_available

    def initialize(self):
        if self._reflectivity_data.utilized_pulses == 0:
            raise self._reflectivity_data.UtilizedNumberOfPulsesBelowMinimumAcceptable()
        self._delay_data.setup_scan_range(self._io_service.get(self._delay_data.setpoint_address).data)

    def update_reflectivity_data(self):
        adc_trace_of_laser_pulses = self._io_service.get(self._reflectivity_data.adc_address).data
        number_of_fel_pulses = self._io_service.get(self._reflectivity_data.bunch_number_address).data
        self._reflectivity_data.update_pulses(adc_trace_of_laser_pulses, number_of_fel_pulses)

    def update(self):
        self._update_delay_setpoint()
        self._update_and_process_data()

    def _update_delay_setpoint(self):
        self._set_delay()
        self._check_delay()

    def _set_delay(self):
        if self._delay_data.is_ready_for_new_setpoint():
            self._io_service.set(self._delay_data.setpoint_address, self._delay_data.get_next_setpoint())

    def _check_delay(self):
        if not self._delay_data.is_ready_for_new_data:
            self._delay_data.check_delay(self._io_service.get(self._delay_data.readback_address).data,
                                         self._io_service.get(self._delay_data.setpoint_address).data)

    def _update_and_process_data(self):
        self._reflectivity_readout = self._io_service.get(self._reflectivity_data.adc_address)
        self._is_new_data_available = False
        if self._reflectivity_readout.event != self._event:
            self._event = self._reflectivity_readout.event
            self._update_data()
            self._process_data()
            self._is_new_data_available = True

    def _update_data(self):
        if self._delay_data.is_ready_for_new_data and self._to_update():
            self._delay_data.update(self._io_service.get(self._delay_data.encoder_address).data)
            self._reflectivity_data.update(self._reflectivity_readout.data)

    def _to_update(self):
        if self._reflectivity_data.is_fast_shutter_used():
            event_odd = bool(self._event % 2)
            if (not self._last_event_odd and event_odd) or (self._last_event_odd and not event_odd):
                self._last_event_odd = event_odd
                return True
            return False
        return True

    def _process_data(self):
        if self._delay_data.is_ready_for_new_scan_point:
            self._delay_data.process_with(self._process_function)
            self._reflectivity_data.process_with(self._process_function)

    def _process_function(self, data):
        indices, bins = self._delay_data.get_indices_and_bins()
        histogram = [[] for _ in range(bins)]
        for index, element in enumerate(data):
            histogram[indices[index] - 1].append(element)
        return np.array([(np.nanmean(h), np.nanstd(h) / np.sqrt(len(h))) for h in histogram])
