from collections import namedtuple

import numpy as np
from doocspie.pyqt import messages

from .alignment_tool_ui import AlignmentToolUi


class AlignmentTool:
    _Fit = namedtuple("Fit", ["position", "metrics", "legend"])

    def __init__(self, alignment_position_data, metrics_data, fit_service):
        self._alignment_tool_ui = AlignmentToolUi()
        self._alignment_position_data = alignment_position_data
        self._metrics_data = metrics_data
        self._fit_service = fit_service
        self._is_started = False
        self._create_connections()

    def _create_connections(self):
        self._alignment_tool_ui.start_button.clicked.connect(self._start_button_clicked)
        self._alignment_tool_ui.stop_button.clicked.connect(self.stop_button_clicked)

    @property
    def ui(self):
        return self._alignment_tool_ui

    @property
    def is_started(self):
        return self._is_started

    def _start_button_clicked(self):
        if self._is_input_validation_successful():
            self._is_started = True
            self._alignment_tool_ui.start_button.setText("running ...")
            self._enable_control_and_input_fields(False)
            self._alignment_position_data.initialize(scan_range=self._get_scan_range(),
                                                     samples_per_step=int(
                                                         self._alignment_tool_ui.samples_per_step_line_edit.text()),
                                                     tolerance=int(
                                                         self._alignment_tool_ui.tolerance_line_edit.text()) / 1000,
                                                     timeout=float(self._alignment_tool_ui.timeout_line_edit.text()))
            self._metrics_data.initialize_alignment_metrics()
            self._alignment_tool_ui.clear()

    def _is_input_validation_successful(self):
        try:
            float(self._alignment_tool_ui.begin_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'Begin' field")
            return False
        try:
            float(self._alignment_tool_ui.end_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'End' field")
            return False
        try:
            int(self._alignment_tool_ui.number_of_steps_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Number of Steps' field")
            return False
        try:
            int(self._alignment_tool_ui.samples_per_step_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Samples / Step' field")
            return False
        try:
            int(self._alignment_tool_ui.tolerance_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Tolerance' field")
            return False
        try:
            float(self._alignment_tool_ui.timeout_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'Timeout' field")
            return False
        return True

    def _enable_control_and_input_fields(self, state):
        self._alignment_tool_ui.start_button.setEnabled(state)
        self._alignment_tool_ui.begin_line_edit.setEnabled(state)
        self._alignment_tool_ui.end_line_edit.setEnabled(state)
        self._alignment_tool_ui.number_of_steps_line_edit.setEnabled(state)
        self._alignment_tool_ui.samples_per_step_line_edit.setEnabled(state)
        self._alignment_tool_ui.tolerance_line_edit.setEnabled(state)
        self._alignment_tool_ui.timeout_line_edit.setEnabled(state)

    def _get_scan_range(self):
        position_range_begin = float(self._alignment_tool_ui.begin_line_edit.text())
        position_range_end = float(self._alignment_tool_ui.end_line_edit.text())
        steps = int(self._alignment_tool_ui.number_of_steps_line_edit.text())
        return np.linspace(position_range_begin, position_range_end, steps, dtype=np.float32)

    def stop_button_clicked(self):
        if self._is_started:
            self._alignment_tool_ui.set_progress(0)
            self._alignment_tool_ui.start_button.setText(AlignmentToolUi.START)
            self._enable_control_and_input_fields(True)
            self._alignment_position_data.reset()
            self._is_started = False

    def update_scanning_progress(self):
        if self._is_started:
            self._alignment_tool_ui.set_progress(self._alignment_position_data.get_scanning_progress())
            if self._alignment_position_data.get_scanning_progress() == 100:
                self._alignment_tool_ui.stop_button.click()

    def update(self):
        if self._alignment_position_data.is_ready_for_new_scan_point:
            self._plot()

    def _plot(self):
        self._alignment_tool_ui.plot_scan(self._alignment_position_data.position,
                                          self._metrics_data.alignment_position_metrics_values,
                                          self._metrics_data.alignment_position_metrics_errors)
        self._plot_fit()

    def _plot_fit(self):
        fit, legend = self._fit_service.get_alignment_fit(self._alignment_position_data.position.values,
                                                          np.array(
                                                              self._metrics_data.alignment_position_metrics_values),
                                                          np.array(
                                                              self._metrics_data.alignment_position_metrics_errors))

        if fit is not None and legend is not None:
            self._alignment_tool_ui.plot_fit(self._Fit(fit[0], fit[1], legend))
        else:
            self._alignment_tool_ui.clear_fit()

    def update_scan(self):
        if self._alignment_position_data.is_not_empty():
            self._plot()
