from doocspie.pyqt import messages

from .profiler_pulse_settings_tool_ui import ProfilerPulseSettingsToolUi


class ProfilerPulseSettingsTool:

    def __init__(self, profiler_service):
        self._profiler_pulse_settings_tool_ui = ProfilerPulseSettingsToolUi()
        self._profiler_pulse_settings_tool_ui.closeEvent = self._close_event
        self._profiler_service = profiler_service
        self._is_initialized = False
        self._create_connections()

    def _close_event(self, event):
        try:
            self._profiler_service.transmission_data.set_parameters(actual_pulse=self._get_actual_pulse(),
                                                                    signal_range=self._get_signal_range(),
                                                                    background_range=self._get_background_range())
        except ValueError:
            event.ignore()

    def _get_actual_pulse(self):
        try:
            return int(self._profiler_pulse_settings_tool_ui.actual_pulse_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Actual 'Pulse' field")
            raise err

    def _get_signal_range(self):
        try:
            begin = int(self._profiler_pulse_settings_tool_ui.signal_begin_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Signal 'Begin' field")
            raise err
        try:
            end = int(self._profiler_pulse_settings_tool_ui.signal_end_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Signal 'End' field")
            raise err

        if begin > end:
            messages.show_error("Signal 'Begin' cannot be larger then 'End'")
            raise ValueError

        return begin, end

    def _get_background_range(self):
        try:
            begin = int(self._profiler_pulse_settings_tool_ui.background_begin_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Background 'Begin' field")
            raise err
        try:
            end = int(self._profiler_pulse_settings_tool_ui.background_end_line_edit.text())
        except ValueError as err:
            messages.show_error("Empty/Faulty Background 'End' field")
            raise err

        if begin > end:
            messages.show_error("Background 'Begin' cannot be larger then 'End'")
            raise ValueError

        return begin, end

    def _create_connections(self):
        self._profiler_pulse_settings_tool_ui.threshold_slider.valueChanged.connect(self._update_threshold)
        self._profiler_pulse_settings_tool_ui.actual_pulse_line_edit.textEdited.connect(self._plot_actual_pulse)

    @property
    def ui(self):
        return self._profiler_pulse_settings_tool_ui

    def _update_threshold(self):
        self._setup_transmission_data()
        self._plot_and_set_pulses()

    def initialize(self):
        if not self._is_initialized:
            self._profiler_service.update_transmission_data()
            self._setup_transmission_data()

    def _setup_transmission_data(self):
        if not self._is_initialized:
            self._profiler_pulse_settings_tool_ui.threshold_slider.setValue(5)
            self._profiler_service.transmission_data.set_parameters(actual_pulse=self._get_actual_pulse(),
                                                                    signal_range=self._get_signal_range(),
                                                                    background_range=self._get_background_range())
        self._profiler_service.transmission_data.find_threshold(
            self._profiler_pulse_settings_tool_ui.threshold_slider.value())
        self._profiler_pulse_settings_tool_ui.threshold_slider.setMinimum(
            self._profiler_service.transmission_data.minimum_threshold_factor)
        self._profiler_pulse_settings_tool_ui.threshold_slider.setMaximum(
            self._profiler_service.transmission_data.maximum_threshold_factor)
        self._profiler_service.transmission_data.find_pulses()
        self._is_initialized = True

    def show_dialog(self):
        self.update()
        self._profiler_pulse_settings_tool_ui.show_window()

    def update(self):
        self._profiler_service.update_transmission_data()
        self._setup_transmission_data()
        self._plot_and_set_pulses()

    def _plot_and_set_pulses(self):
        self._plot()
        self._set_pulses()

    def _plot(self):
        self._profiler_pulse_settings_tool_ui.plot_adc(self._profiler_service.transmission_data.adc)
        self._profiler_pulse_settings_tool_ui.plot_threshold(self._profiler_service.transmission_data.threshold)
        self._profiler_pulse_settings_tool_ui.plot_positions(self._profiler_service.transmission_data.positions)
        self._plot_actual_pulse()

    def _plot_actual_pulse(self):
        actual_pulse = self._profiler_pulse_settings_tool_ui.actual_pulse_line_edit.text()
        if actual_pulse:
            actual_pulse_position = int(actual_pulse) - 1
            try:
                position = self._profiler_service.transmission_data.positions.bins[actual_pulse_position]
                value = self._profiler_service.transmission_data.positions.values[actual_pulse_position]
            except IndexError:
                messages.show_error("Actual 'Pulse' out of range ... set to maximum pulse")
                self._profiler_pulse_settings_tool_ui.actual_pulse_line_edit.setText(
                    str(self._profiler_service.transmission_data.pulses))
                self._plot_actual_pulse()
                return
            self._profiler_pulse_settings_tool_ui.plot_actual_pulse(position, value)

    def _set_pulses(self):
        self._profiler_pulse_settings_tool_ui.set_detected_pulses(self._profiler_service.transmission_data.pulses)
