from collections import namedtuple

import numpy as np
from doocspie.pyqt import messages

from .profiler_tool_ui import ProfilerToolUi


class ProfilerTool:
    _Fit = namedtuple("Fit", ["position", "transmission", "legend"])

    def __init__(self, profiler_service, fit_service):
        self._profiler_tool_ui = ProfilerToolUi()
        self._profiler_service = profiler_service
        self._profiler_positions = self._profiler_service.profiler_positions
        self._profiler_position_data = None
        self._transmission_data = self._profiler_service.transmission_data
        self._fit_service = fit_service
        self._fit = None
        self._legend = None
        self._plot_label = None
        self._scan_dimension = None
        self._is_started = False
        self._create_connections()
        self._update_scan_dimension()

    def _create_connections(self):
        self._profiler_tool_ui.start_button.clicked.connect(self._start_button_clicked)
        self._profiler_tool_ui.stop_button.clicked.connect(self._stop_button_clicked)

    @property
    def ui(self):
        return self._profiler_tool_ui

    @property
    def is_started(self):
        return self._is_started

    def _update_scan_dimension(self):
        if self._profiler_tool_ui.horizontal_radio_button.isChecked():
            self._profiler_position_data = self._profiler_positions.horizontal
            self._plot_label = "Horizontal Readback Position (mm)"
            self._scan_dimension = "horizontal"
        elif self._profiler_tool_ui.vertical_radio_button.isChecked():
            self._profiler_position_data = self._profiler_positions.vertical
            self._plot_label = "Vertical Readback Position (mm)"
            self._scan_dimension = "vertical"
        self._profiler_service.set_data(self._profiler_position_data)

    def _start_button_clicked(self):
        if self._is_input_validation_successful():
            self._is_started = True
            self._profiler_tool_ui.start_button.setText("running ...")
            self._enable_control_and_input_fields(False)
            self._update_scan_dimension()
            self._profiler_position_data.initialize(scan_range=self._get_scan_range(),
                                                    samples_per_step=int(
                                                        self._profiler_tool_ui.samples_per_step_line_edit.text()),
                                                    tolerance=int(
                                                        self._profiler_tool_ui.tolerance_line_edit.text()) / 1000,
                                                    timeout=float(self._profiler_tool_ui.timeout_line_edit.text()))
            self._transmission_data.initialize()
            self._profiler_tool_ui.clear()

    def _is_input_validation_successful(self):
        try:
            float(self._profiler_tool_ui.begin_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'Begin' field")
            return False
        try:
            float(self._profiler_tool_ui.end_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'End' field")
            return False
        try:
            int(self._profiler_tool_ui.number_of_steps_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Number of Steps' field")
            return False
        try:
            int(self._profiler_tool_ui.samples_per_step_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Samples / Step' field")
            return False
        try:
            int(self._profiler_tool_ui.tolerance_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Tolerance' field")
            return False
        try:
            float(self._profiler_tool_ui.timeout_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'Timeout' field")
            return False
        return True

    def _enable_control_and_input_fields(self, state):
        self._profiler_tool_ui.start_button.setEnabled(state)
        self._profiler_tool_ui.horizontal_radio_button.setEnabled(state)
        self._profiler_tool_ui.vertical_radio_button.setEnabled(state)
        self._profiler_tool_ui.begin_line_edit.setEnabled(state)
        self._profiler_tool_ui.end_line_edit.setEnabled(state)
        self._profiler_tool_ui.number_of_steps_line_edit.setEnabled(state)
        self._profiler_tool_ui.samples_per_step_line_edit.setEnabled(state)
        self._profiler_tool_ui.tolerance_line_edit.setEnabled(state)
        self._profiler_tool_ui.timeout_line_edit.setEnabled(state)
        self._profiler_tool_ui.check_button.setEnabled(state)

    def _get_scan_range(self):
        position_range_begin = float(self._profiler_tool_ui.begin_line_edit.text())
        position_range_end = float(self._profiler_tool_ui.end_line_edit.text())
        steps = int(self._profiler_tool_ui.number_of_steps_line_edit.text())
        return np.linspace(position_range_begin, position_range_end, steps, dtype=np.float32)

    def _stop_button_clicked(self):
        if self._is_started:
            self._profiler_tool_ui.set_progress(0)
            self._profiler_tool_ui.start_button.setText(ProfilerToolUi.START)
            self._enable_control_and_input_fields(True)
            self._profiler_position_data.reset()
            self._is_started = False

    def update_profiler_progress(self):
        if self._is_started:
            self._profiler_tool_ui.set_progress(self._profiler_position_data.get_scanning_progress())
            if self._profiler_position_data.get_scanning_progress() == 100:
                self._profiler_tool_ui.stop_button.click()

    def update(self):
        if self._profiler_position_data.is_ready_for_new_scan_point:
            self._plot()

    def _plot(self):
        self._profiler_tool_ui.plot_scan(self._profiler_position_data.position, self._transmission_data.transmission,
                                         self._plot_label)
        self._plot_fit()

    def _plot_fit(self):
        self._fit, self._legend = self._fit_service.get_profiler_fit(
            self._profiler_position_data.position, self._transmission_data.transmission)

        if self._fit is not None and self._legend is not None:
            self._profiler_tool_ui.plot_fit(self._Fit(self._fit[0], self._fit[1], self._legend))
        else:
            self._profiler_tool_ui.clear_fit()

    def update_scan(self):
        if self._profiler_position_data.is_not_empty():
            self._plot()

    def get_export_content(self):
        position_mm = "position (mm)"
        position_error_mm = "position error (mm)"
        transmission_percent = "transmission (%)"
        transmission_error_percent = "transmission error (%)"
        fit_position_mm = "fit position (mm)"
        fit_transmission_percent = "fit transmission (%)"
        fit_legend = "fit legend"

        content = {position_mm: [], position_error_mm: [], transmission_percent: [], transmission_error_percent: [],
                   fit_position_mm: [], fit_transmission_percent: [], fit_legend: [],
                   "scan dimension": [self._scan_dimension],
                   "position begin (mm)": [self._profiler_tool_ui.begin_line_edit.text()],
                   "position end (mm)": [self._profiler_tool_ui.end_line_edit.text()],
                   "number of steps": [self._profiler_tool_ui.number_of_steps_line_edit.text()],
                   "samples per step": [self._profiler_tool_ui.samples_per_step_line_edit.text()],
                   "tolerance (\u00B5m)": [self._profiler_tool_ui.tolerance_line_edit.text()],
                   "timeout (s)": [self._profiler_tool_ui.timeout_line_edit.text()],
                   "actual pulse": [self._profiler_service.transmission_data.actual_pulse]}

        if self._profiler_position_data.is_not_empty():
            content[position_mm] = self._profiler_position_data.position.values
            content[position_error_mm] = self._profiler_position_data.position.errors
            content[transmission_percent] = self._transmission_data.transmission.values
            content[transmission_error_percent] = self._transmission_data.transmission.errors
            if self._fit is not None:
                content[fit_position_mm] = self._fit[0]
                content[fit_transmission_percent] = self._fit[1]
            if self._legend is not None:
                content[fit_legend] = [self._legend.replace("&#177;", "\u00b1")]  # html -> unicode for "+-"

        return content
