from collections import namedtuple

import numpy as np
from doocspie.pyqt import messages

from .timing_tool_ui import TimingToolUi


class TimingTool:
    _Fit = namedtuple("Fit", ["delay", "reflectivity", "legend"])

    def __init__(self, delay_data, reflectivity_data, fit_service):
        self._timing_tool_ui = TimingToolUi()
        self._delay_data = delay_data
        self._reflectivity_data = reflectivity_data
        self._reflectivity_data.use_fast_shutter(self._timing_tool_ui.fast_shutter_check_box.isChecked())
        self._fit_service = fit_service
        self._fit = None
        self._legend = None
        self._is_started = False
        self._create_connections()

    def _create_connections(self):
        self._timing_tool_ui.start_button.clicked.connect(self._start_button_clicked)
        self._timing_tool_ui.stop_button.clicked.connect(self._stop_button_clicked)
        self._timing_tool_ui.fast_shutter_check_box.toggled.connect(self._reflectivity_data.use_fast_shutter)

    @property
    def ui(self):
        return self._timing_tool_ui

    @property
    def is_started(self):
        return self._is_started

    def _start_button_clicked(self):
        if self._is_input_validation_successful():
            self._is_started = True
            self._timing_tool_ui.start_button.setText("running ...")
            self._enable_control_and_input_fields(False)
            self._delay_data.initialize(scan_range=self._get_scan_range(),
                                        samples_per_step=self._get_effective_samples_per_step())
            self._reflectivity_data.initialize()
            self._timing_tool_ui.clear()

    def _is_input_validation_successful(self):
        try:
            float(self._timing_tool_ui.begin_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'Begin' field")
            return False
        try:
            float(self._timing_tool_ui.end_line_edit.text())
        except ValueError:
            messages.show_error("Empty/Faulty 'End' field")
            return False
        try:
            int(self._timing_tool_ui.step_size_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Step Size' field")
            return False
        try:
            int(self._timing_tool_ui.samples_per_step_line_edit.text())
        except ValueError:
            messages.show_error("Empty 'Samples / Step' field")
            return False
        if float(self._timing_tool_ui.begin_line_edit.text()) == float(self._timing_tool_ui.end_line_edit.text()):
            messages.show_error("Delay Range 'Begin' and 'End' cannot be equal")
            return False
        return True

    def _get_effective_samples_per_step(self):
        samples_per_step = int(self._timing_tool_ui.samples_per_step_line_edit.text())
        if self._timing_tool_ui.fast_shutter_check_box.isChecked():
            return 2 * samples_per_step
        return samples_per_step

    def _get_scan_range(self):
        delay_begin = float(self._timing_tool_ui.begin_line_edit.text())
        delay_end = float(self._timing_tool_ui.end_line_edit.text())
        step_size = int(self._timing_tool_ui.step_size_line_edit.text()) / 1000
        steps = np.ceil(abs(delay_end - delay_begin) / step_size + 1).astype(int)
        return np.linspace(delay_begin, delay_end, steps)

    def _enable_control_and_input_fields(self, state):
        self._timing_tool_ui.start_button.setEnabled(state)
        self._timing_tool_ui.begin_line_edit.setEnabled(state)
        self._timing_tool_ui.end_line_edit.setEnabled(state)
        self._timing_tool_ui.step_size_line_edit.setEnabled(state)
        self._timing_tool_ui.samples_per_step_line_edit.setEnabled(state)
        self._timing_tool_ui.fast_shutter_check_box.setEnabled(state)
        self._timing_tool_ui.check_button.setEnabled(state)

    def _stop_button_clicked(self):
        if self._is_started:
            self._timing_tool_ui.set_progress(0)
            self._timing_tool_ui.start_button.setText(TimingToolUi.START)
            self._enable_control_and_input_fields(True)
            self._delay_data.reset()
            self._is_started = False

    def update_timing_progress(self):
        if self._is_started:
            self._timing_tool_ui.set_progress(self._delay_data.get_scanning_progress())

    def update(self):
        if self._delay_data.is_ready_for_new_scan_point:
            self._plot()

    def _plot(self):
        self._timing_tool_ui.plot_timing(self._delay_data.delay, self._reflectivity_data.reflectivity)
        self._plot_fit()

    def _plot_fit(self):
        self._fit, self._legend = self._fit_service.get_timing_fit(
            self._delay_data.delay, self._reflectivity_data.reflectivity)

        if self._fit is not None and self._legend is not None:
            self._timing_tool_ui.plot_fit(self._Fit(self._fit[0], self._fit[1], self._legend))
        else:
            self._timing_tool_ui.clear_fit()

    def update_scan(self):
        if self._reflectivity_data.is_not_empty():
            self._plot()

    def get_export_content(self):
        delay_ps = "delay (ps)"
        delay_error_ps = "delay error (ps)"
        reflectivity_percent = "reflectivity (%)"
        reflectivity_error_percent = "reflectivity error (%)"
        fit_delay_ps = "fit delay (ps)"
        fit_reflectivity_percent = "fit reflectivity (%)"
        fit_legend = "fit legend"

        content = {delay_ps: [], delay_error_ps: [], reflectivity_percent: [], reflectivity_error_percent: [],
                   fit_delay_ps: [], fit_reflectivity_percent: [], fit_legend: [],
                   "delay begin (ps)": [self._timing_tool_ui.begin_line_edit.text()],
                   "delay end (ps)": [self._timing_tool_ui.end_line_edit.text()],
                   "step size (fs)": [self._timing_tool_ui.step_size_line_edit.text()],
                   "samples per step": [self._timing_tool_ui.samples_per_step_line_edit.text()]}

        if self._reflectivity_data.is_not_empty():
            content[delay_ps] = self._delay_data.delay.values
            content[delay_error_ps] = self._delay_data.delay.errors
            content[reflectivity_percent] = self._reflectivity_data.reflectivity.values
            content[reflectivity_error_percent] = self._reflectivity_data.reflectivity.errors
            if self._fit is not None:
                content[fit_delay_ps] = self._fit[0]
                content[fit_reflectivity_percent] = self._fit[1]
            if self._legend is not None:
                content[fit_legend] = [self._legend]

        return content
